{-# LANGUAGE BangPatterns #-}
{-# LANGUAGE DeriveGeneric #-}
{-# LANGUAGE ExistentialQuantification #-}
{-# OPTIONS -Wall #-}

module Main where

import qualified Data.ByteString.Lazy.Char8 as BS
import System.Environment
import System.Random
import GHC.Generics
import Data.Aeson

import StrategyManager
import Strategy0
import Datatypes.Game(Game,Command,commandsToString,stringToCommands)
--import VM
import Opt
import JSONDeser(readInput)
import ThreadPoolTimed

strategyTag :: String
strategyTag = "lilik0"

logFileName :: String
logFileName = "scores" 

timeLimitRatio :: Double
timeLimitRatio = 0.96

memLimitRatio :: Double
memLimitRatio = 1.0

computationsPerStep :: Int
computationsPerStep = 100

data JSONSer = JSONSer { id :: Int,
                         seed :: Int,
                         tag :: String,
                         problemSolution :: String
                       } deriving (Show, Generic)

instance FromJSON JSONSer
instance ToJSON JSONSer

type Id = Int
type Seed = Int


strategies :: Game -> StdGen ->  [[Command]] -> GameComputation
strategies g sgen cmd = [MkStrategyWrapper (initst g sgen cmd :: Strategy0)]

-- = [MkStrategyWrapper (initst g sgen cmd :: NullStrategy1),
--    MkStrategyWrapper (initst g sgen cmd :: NullStrategy2)]

-- example ::
-- = [MkStrategyWrapper (init g sgen cmd :: Strat0),
--    MkStrategyWrapper (init g sgen cmd :: Strat1),
--    MkStrategyWrapper (init g sgen cmd :: Strat2)]

main :: IO ()
main = do args <- getArgs
          opt <- parseArgs args
          let files       = optFile                opt
          let maxTime     = optTime                opt
--          let maxMem      = optMem                 opt
          let powerPhrases = optPowerPhrase        opt
          let logFile     = optLog                 opt
          let cores       = optCores               opt
          rng <- getStdGen
          initialData <- createComputationsFromFiles files rng powerPhrases
          let (_, _,gameComputations) = unzip3 initialData
          commandResults <- iterateGame gameComputations cores  maxTime
          let stringResults = map (\(cmds,score,algoIdx) -> (commandsToString cmds,score,algoIdx)) commandResults
          let outJSONstructs = zipWith jsonBuilder initialData stringResults
          BS.putStrLn $ encode outJSONstructs
          writeLogFile logFile (zipWith logFileBuilder initialData stringResults)
  where
--    timeStruct Nothing    _               = Nothing
--    timeStruct (Just stopTime) initialTime = Just (fromIntegral stopTime,fromIntegral initialTime)
                                            
                                                        
                                                         
    jsonBuilder (idx, seedd, _) (strCmds, _, _)         = (JSONSer idx seedd strategyTag strCmds) 
    logFileBuilder (idx, seedd, _) (_ ,score , algoIdx) = (idx, seedd, score, algoIdx)
                                     
createComputationsFromFiles :: [String] -> StdGen -> [String] -> IO [(Id,Seed,GameComputation)]
createComputationsFromFiles fileNames randomGen powerPhrases = do inputs <- readFiles fileNames
                                                                  let igames  = map readInput inputs 
                                                                  let cstruct = compstruct igames
                                                                  return (gcstruct cstruct)
  where
    compstruct ig = concat (map genf ig)
    genf    (i,g) = zipWith (\x (y,z) -> (x,y,z)) (replicate (length g) i) g
    gcstruct cst  = map (\(x,y,z) -> (x,y,j z)) cst
      where
        j game = strategies game randomGen (map stringToCommands powerPhrases)

readFiles :: [String] -> IO [BS.ByteString]
readFiles []     = return []
readFiles (x:xs) = do f  <- BS.readFile x
                      fs <- readFiles xs
                      return (f:fs)


iterateGame :: [GameComputation] -> Int -> Int -> IO [FinishedGame] 
iterateGame gameComputations cores timeLimit = do results  <- threadPoolTimed cores timeLimit advanceGameComputation gameComputations
                                                  return (map getBestGameComputation results)
  where
    

writeLogFile :: Bool -> [(Int,Int,Int,Int)] -> IO ()
writeLogFile False _  = return ()
writeLogFile _ els    = writeFile logFileName scoredata
  where
    scoredata = foldl strlog "\n" els
    strlog x (a,b,c,d) = sa ++ sb ++ sc ++ sd ++ x ++ "\n\n" 
      where
        sa = (show a) ++ " "
        sb = (show b) ++ " "
        sc = (show c) ++ " "
        sd = (show d) ++ " "

applyNtimes :: Int -> (a -> a) -> a -> a
applyNtimes 0 _ accum = accum
applyNtimes n f accum = applyNtimes (n - 1) f (f accum)
    
