{-# OPTIONS -Wall #-}
module MonadParParallelization(parallelizedTimed) where

import Control.DeepSeq (deepseq)
import Control.Concurrent.STM
import Control.Concurrent
import Control.Monad.Par


writeTVarIO :: TVar a -> a -> IO ()
writeTVarIO tvar el = atomically $ writeTVar tvar el 

parallelizedTimed :: (NFData a) => Int -> (a -> a) -> [a] -> IO [a]
parallelizedTimed sec f toCompute = do stopVar <- newTVarIO False
                                       putStrLn "WORKERS STARTED"
                                       worker <- monadParFork (workerFunction f stopVar) toCompute
                                       putStrLn "CLOCK STARTED"
                                       secWait sec
                                       putStrLn "STOP MESSAGE SENT"
                                       sendStopMessage stopVar
                                       putStrLn "WAITING FOR COMPLETITION"
                                       waitWorkerCompletition worker
                                       putStrLn "RETRIEVE DATA"
                                       retrieveAllData worker
  where
    secWait = threadDelay . (1000000 * )
    sendStopMessage stopMessageC = writeTVarIO stopMessageC True

monadParFork :: ((TVar Bool,TVar [a], [a]) -> IO ()) -> [a] -> IO (TVar Bool,TVar [a])
monadParFork f toCompute = do stopThreadVar <- newTVarIO False
                              dataTVar <- newTVarIO toCompute
                              _ <- forkIO (f (stopThreadVar, dataTVar, toCompute))
                              return (stopThreadVar, dataTVar)
    
workerFunction :: (NFData a) => (a -> a) -> TVar Bool -> (TVar Bool, TVar [a], [a]) -> IO ()
workerFunction f stopVar (threadEndedVar, dataVar, dataEls) = do stopped <- readTVarIO stopVar
                                                                 if stopped
                                                                 then do writeTVarIO dataVar dataEls
                                                                         writeTVarIO threadEndedVar True
                                                                 else do let newData = runPar evalf
                                                                         newData `deepseq` workerFunction f stopVar (threadEndedVar, dataVar, newData)
  where
    evalf = do iVars <- mapM (\_ -> new) dataEls
               let iVarsPEls = zip iVars dataEls
               mapM_ (\(ivar, datael) -> fork (put ivar (f datael))) iVarsPEls
               gotDatas <- mapM (\(ivar, _) -> get ivar) iVarsPEls
               return gotDatas

waitWorkerCompletition :: (TVar Bool, TVar [a]) -> IO ()
waitWorkerCompletition workers = do  finished <- readTVarIO (fst workers)
                                     if finished
                                     then return ()
                                     else waitWorkerCompletition workers
  
retrieveAllData :: (TVar Bool, TVar [a]) -> IO [a]
retrieveAllData workers = readTVarIO (snd workers)  
