{-# LANGUAGE DeriveGeneric #-}
module Main where

import GHC.Generics
import Data.Aeson
import Data.Aeson.Types
import qualified Data.List as List

import System.Environment
import qualified Data.ByteString.Lazy.Char8 as BS
import System.IO 
import Datatypes
import Opt
import JSONDeser
import Strategy0
import Datatypes.Game (commandsToString)
import Datatypes.Game (Command)
import PowerPhrases (charToCommand)

import qualified Datatypes.Game as Game
import Datatypes.Game (Game)
import VM (step, Notes(..))

data JSONSer = JSONSer { problemId :: Int,
                         seed :: Int,
                         tag :: String,
                         solution :: String
                       } deriving (Show, Generic)

instance FromJSON JSONSer
instance ToJSON JSONSer

testVM :: Game -> String -> ([Notes], Int, Game)
testVM game strCmds = buildResult game commands [] where
    commands = [c | (Just c) <- (map charToCommand strCmds)]
    buildResult game [] notes = (reverse notes, Game.score game + (Game.powerCounterToScore $ Game.powerCounter game), game)
    buildResult game (c:cs) notes = let (newGame, note) = step game c
                                        in if elem note [ErrorSamePosition, ErrorGameEnded]
                                               then (reverse (note:notes), 0, newGame)
                                               else buildResult newGame cs (note:notes)

main :: IO ()
main = do args <- getArgs
          opt <- parseArgs args
          file <- return ((optFile opt) !! 0)
          seed <- return (optSeedNumber opt)
          (Just commands) <- return (optPowerPhrase opt)
          str <- BS.readFile file
          (_, gmseed) <- return (readInput str)
          let (Just (s, g)) = List.find (\(s, g) -> s == seed) gmseed
          (notes, score, ng) <- return $ testVM g commands
          putStrLn ("Commands: " ++ show (Game.commandsToString . reverse . Game.history $ ng))
          putStrLn ("Notes: " ++ show notes)
          putStrLn ("Score: " ++ show score)
          
packAll :: Int -> [Int] -> [[Command]] -> [JSONSer] 
packAll id seeds commandLists = zipWith (\x y -> JSONSer id x "lilik0" y) seeds commandStrings
  where
    commandStrings = map commandsToString commandLists

scoredata :: Int -> [Int] -> [Int] -> String
scoredata id seeds points = pretty
  where
    pretty = foldl (\x (a,b,c) -> (show a) ++ " " ++ (show b) ++  " " ++ (show c) ++ "\n" ++ x) "" zipdata
    zipdata = zip3 ids seeds points
    ids = replicate (length seeds) id
    
