{-# OPTIONS -Wall #-}
module ThreadPoolTimed(threadPoolTimed) where

import Control.DeepSeq (NFData, deepseq)
import Control.Concurrent.STM.TChan
import Control.Concurrent.STM.TVar
import Control.Monad.STM
import Control.Concurrent
import Control.Monad
import Data.Maybe


threadPoolTimed :: (NFData a) => Int -> Int -> (a -> a) -> [a] -> IO [a]
threadPoolTimed nworkers sec f toCompute = do toComputeQ <- newTChanIO :: IO (TChan a)
                                              stopVar <- newTVarIO False
                                              putStrLn "DATA TO WORKERS"
                                              pushDataToWorkers toCompute toComputeQ
                                              putStrLn "WORKERS STARTED"
                                              workers <- forkNTimes nworkers (workerFunction f (toComputeQ,
                                                                                                stopVar))
                                              putStrLn "CLOCK STARTED"
                                              secWait sec
                                              putStrLn "STOP MESSAGE SENT"
                                              sendStopMessage stopVar
                                              putStrLn "WAITING FOR COMPLETITION"
                                              waitWorkerCompletition workers
                                              putStrLn "RETRIEVE DATA"
                                              retrieveAllData toComputeQ
                                              
  where
    secWait = threadDelay . (1000000 * )
    sendStopMessage stopMessageC = atomically $ writeTVar stopMessageC True

forkNTimes :: Int -> (TVar Bool -> IO ()) -> IO [TVar Bool]
forkNTimes nworkers f = do workers <- replicateM nworkers (newTVarIO False)
                           mapM_ (forkIO . f) workers
                           return workers
                           

workerFunction :: (NFData a) => (a -> a) -> (TChan a, TVar Bool) -> TVar Bool -> IO ()
workerFunction f (toCompute, stopVar) threadEndedVar = do (stopMessageArrived, computationData) <- retrieveData
                                                          if stopMessageArrived
                                                          then atomically $ do writeTVar threadEndedVar True                              
                                                          else do executeStep computationData
                                                                  workerFunction f (toCompute, stopVar) threadEndedVar
  where
    retrieveData = atomically $ do stopMessage <- readTVar stopVar
                                   if stopMessage
                                   then return (True, Nothing)
                                   else do compValue <- readTChan toCompute
                                           return (False, Just compValue) 

    executeStep computationData = do let newData = ((f . fromJust) computationData)
                                     newData `deepseq` atomically $ writeTChan toCompute newData


pushDataToWorkers :: [a] -> TChan a -> IO ()
pushDataToWorkers toCompute toComputeQ = atomically $ mapM_ (\el -> writeTChan toComputeQ el) toCompute



--retrieveData :: TChan a -> TChan b -> [(a,b)]
--retrieveData unCompletedQ completedQ

waitWorkerCompletition :: [TVar Bool] -> IO ()
waitWorkerCompletition workers = atomically $ do workersStatus <- mapM readTVar workers
                                                 if and workersStatus
                                                 then return ()
                                                 else retry


retrieveAllData :: TChan a -> IO [a]

retrieveAllData queue = atomically $ retrieveAllDataSTM []
  where retrieveAllDataSTM acc = do isEmpty <- isEmptyTChan queue
                                    if isEmpty
                                    then return acc
                                    else do elemToAcc <- readTChan queue
                                            retrieveAllDataSTM (elemToAcc:acc)

-- retrieveAllData queue acc = do voidQueue <- atomically $ isEmptyTChan queue
--                                putStrLn $ (show voidQueue)
--                                if voidQueue
--                                then return acc
--                                else do elemToAcc <- atomically $ readTChan queue
--                                        retrieveAllData queue (elemToAcc:acc)
                                         
