{-# OPTIONS -Wall #-}
module ThreadPoolTimed2(threadPoolTimed2) where

import Control.DeepSeq (NFData, deepseq)
import Control.Concurrent.STM
import Control.Concurrent

writeTVarIO :: TVar a -> a -> IO ()
writeTVarIO tvar el = atomically $ writeTVar tvar el 

threadPoolTimed2 :: (NFData a) => Int -> (a -> a) -> [a] -> IO [a]
threadPoolTimed2 sec f toCompute = do stopVar <- newTVarIO False
                                      putStrLn "WORKERS STARTED"
                                      workers <- mapFork (workerFunction f stopVar) toCompute
                                      putStrLn "CLOCK STARTED"
                                      secWait sec
                                      putStrLn "STOP MESSAGE SENT"
                                      sendStopMessage stopVar
                                      putStrLn "WAITING FOR COMPLETITION"
                                      waitWorkerCompletition workers
                                      putStrLn "RETRIEVE DATA"
                                      retrieveAllData workers
  where
    secWait = threadDelay . (1000000 * )
    sendStopMessage stopMessageC = writeTVarIO stopMessageC True

mapFork :: ((TVar Bool, TVar a, a) -> IO ()) -> [a] -> IO [(TVar Bool,TVar a, a)]
mapFork f toCompute = do workers <- mapM tupleM toCompute
                         mapM_ (forkIO . f) workers
                         return workers
  where
    tupleM x = do falseTVar <- newTVarIO False
                  dataTvar <- newTVarIO x
                  return (falseTVar, dataTvar, x)
    
workerFunction :: (NFData a) => (a -> a) -> TVar Bool -> (TVar Bool, TVar a, a) -> IO ()
workerFunction f stopVar (threadEndedVar, dataVar, dataEl) = do stopped <- readTVarIO stopVar
                                                                if stopped
                                                                then do writeTVarIO dataVar dataEl
                                                                        writeTVarIO threadEndedVar True
                                                                else do let newData = dataEl `deepseq` f(dataEl)
                                                                        newData `deepseq` workerFunction f stopVar (threadEndedVar, dataVar, newData)



waitWorkerCompletition :: [(TVar Bool, TVar a, a)] -> IO ()
waitWorkerCompletition workers = do workerFinishedList <- mapM (\(x, _, _) -> readTVarIO x) workers
                                    let workersAllFinished = and workerFinishedList
                                    if workersAllFinished
                                    then return ()
                                    else waitWorkerCompletition workers

retrieveAllData :: [(TVar Bool, TVar a, a)] -> IO [a]
retrieveAllData workers = mapM (\(_, x, _) -> readTVarIO x) workers 
