{-# OPTIONS -Wall #-}
module MonadParParallelization(parallelizedTimed) where

import Control.DeepSeq (deepseq)
import Control.Concurrent.MVar
import Control.Concurrent
import Control.Monad.Par
  
parallelizedTimed :: (NFData a) => Int -> (a -> a) -> [a] -> IO [a]
parallelizedTimed sec f toCompute = do stopVar <- newMVar False
                                       putStrLn "WORKERS STARTED"
                                       worker <- monadParFork (workerFunction f stopVar) toCompute
                                       putStrLn "CLOCK STARTED"
                                       secWait sec
                                       putStrLn "STOP MESSAGE SENT"
                                       sendStopMessage stopVar
                                       putStrLn "WAITING FOR COMPLETITION"
                                       waitWorkerCompletition worker
                                       putStrLn "RETRIEVE DATA"
                                       retrieveAllData worker
  where
    secWait = threadDelay . (1000000 * )
    sendStopMessage stopMessageC = putMVar stopMessageC True

monadParFork :: ((MVar Bool,MVar [a], [a]) -> IO ()) -> [a] -> IO (MVar Bool,MVar [a])
monadParFork f toCompute = do stopThreadVar <- newMVar False
                              dataMVar <- newMVar toCompute
                              _ <- forkIO (f (stopThreadVar, dataMVar, toCompute))
                              return (stopThreadVar, dataMVar)
    
workerFunction :: (NFData a) => (a -> a) -> MVar Bool -> (MVar Bool, MVar [a], [a]) -> IO ()
workerFunction f stopVar (threadEndedVar, dataVar, dataEls) = do stopped <- readMVar stopVar
                                                                 if stopped
                                                                   then do putMVar dataVar dataEls
                                                                           putMVar threadEndedVar True
                                                                   else do let newData = evalf
                                                                           newData `deepseq` workerFunction f stopVar (threadEndedVar, dataVar, newData)
  where
    evalf = parMap `deepseq` f dataEls  

waitWorkerCompletition :: (MVar Bool, MVar [a]) -> IO ()
waitWorkerCompletition workers = do  finished <- readMVar (fst workers)
                                     if finished
                                     then return ()
                                     else waitWorkerCompletition workers
  
retrieveAllData :: (MVar Bool, MVar [a]) -> IO [a]

retrieveAllData workers = readMVar (snd workers)  
